/*
 * Functions for the "list" action.
 *
 * Copyright 2024 Andrew Wood
 *
 * License GPLv3+: GNU GPL version 3 or later; see `docs/COPYING'.
 */

#include "scw-internal.h"
#include <string.h>

/*
 * List all defined items.
 */
int showItemList(struct scwState *state)
{
	char **usernameArray = NULL;
	size_t usernameCount = 0;
	size_t usernameIndex;

	if (state->allUsers) {
		enumerateAllUsers(state, &usernameArray, &usernameCount);
	} else {
		usernameArray = (char **) stringCopy(state, NULL, sizeof(char *) - 1);
		if (NULL == usernameArray)
			return SCW_EXIT_ERROR;
		/*@-dependenttrans@ */
		usernameArray[0] = state->username;
		/*@+dependenttrans@ */
		usernameCount = 1;
	}

	for (usernameIndex = 0; NULL != usernameArray && usernameIndex < usernameCount; usernameIndex++) {
		char *originalUsername;
		size_t originalUsernameLength;
		char *username;
		size_t usernameLength;
		char **itemArray = NULL;
		size_t itemCount = 0;
		size_t itemIndex;

		username = usernameArray[usernameIndex];
		if (NULL == username)
			continue;
		usernameLength = strlen(username);	/* flawfinder: ignore */
		/*
		 * flawfinder warns of strlen() with non-null-terminated
		 * strings; the array contains strings we have explicitly
		 * null terminated.
		 */

		originalUsername = state->username;
		originalUsernameLength = state->usernameLength;
		state->username = username;
		state->usernameLength = usernameLength;

		/* Load the per-user configuration for this user. */
		memset(&(state->userSettings), 0, sizeof(state->userSettings));
		if (0 != loadUserConfig(state)) {
			state->username = originalUsername;
			state->usernameLength = originalUsernameLength;
			continue;
		}

		/* Find the items for this user. */
		enumerateUserItems(state, &itemArray, &itemCount);

		/* Load each item and display its information. */
		for (itemIndex = 0; NULL != itemArray && itemIndex < itemCount; itemIndex++) {
			struct scwSettings combinedSettings;
			bool itemDisabled = false;

			state->item = itemArray[itemIndex];
			if (NULL == state->item)
				continue;
			state->itemLength = strlen(state->item);	/* flawfinder: ignore */
			/*
			 * flawfinder warns of strlen() with
			 * non-null-terminated strings; the array contains
			 * strings we have explicitly null terminated.
			 */

			/* Load the item's settings. */
			memset(&(state->itemSettings), 0, sizeof(state->itemSettings));
			if (0 != loadCurrentItemSettings(state))
				continue;

			/* Combine and expand the settings for this item. */
			memset(&combinedSettings, 0, sizeof(combinedSettings));
			if (0 != combineSettings(state, &combinedSettings))
				continue;
			expandAllRawValues(state, &combinedSettings);

			/* Check whether the item is disabled. */
			if (NULL != combinedSettings.metricsDir.expandedValue) {
				if (fileExistsAt
				    (combinedSettings.metricsDir.expandedValue,
				     combinedSettings.metricsDir.expandedLength, STATIC_STRING("disabled"))) {
					itemDisabled = true;
				}
			}

			/* Skip if disabled and we're only listing enabled items. */
			if (state->listEnabledOnly && itemDisabled)
				continue;

			/* Skip if enabled and we're only listing disabled items. */
			if (state->listDisabledOnly && (!itemDisabled))
				continue;

			/* Display the information. */
			/*@-mustfreefresh@ */
			/* splint note: gettext triggers a warning we can't resolve. */
			printf("%.*s\t%.*s\t%s\t", (int) (state->usernameLength), state->username,
			       (int) (state->itemLength), state->item, itemDisabled ? _("disabled") : _("enabled"));
			if (NULL == combinedSettings.description.expandedValue
			    || combinedSettings.description.expandedLength < 1) {
				printf("%s", _("(no description)"));
			} else {
				printf("%.*s", (int) (combinedSettings.description.expandedLength),
				       combinedSettings.description.expandedValue);
			}
			/*@+mustfreefresh@ */
			printf("\n");
		}

		state->item = NULL;
		state->itemLength = 0;

		state->username = originalUsername;
		state->usernameLength = originalUsernameLength;
	}

	return 0;
}
